import ast


class BinaryOp(object):
    def __init__(self, parent, left, operator, right):
        self.parent = parent
        self.left = left
        self.operator = operator
        self.right = right

    def to_node(self):
        pass


class OrTest(BinaryOp):
    def to_node(self):
        if len(self.right) == 0:
            return self.left.to_node()
        else:
            nodes = [self.left.to_node()]
            for node in self.right:
                nodes.append(node.to_node())
            return ast.BoolOp(op=ast.Or(), values=nodes)


class AndTest(BinaryOp):
    def to_node(self):
        if len(self.right) == 0:
            return self.left.to_node()
        else:
            nodes = [self.left.to_node()]
            for node in self.right:
                nodes.append(node.to_node())
            return ast.BoolOp(ast.And(), nodes)


class Comparison(BinaryOp):
    def to_node(self):
        if len(self.operator) == 0:
            return self.left.to_node()
        else:
            right_nodes = []
            comp_operator = []
            for i in range(len(self.right)):
                right_nodes.append(self.right[i].to_node())
                if self.operator[i] in ['<', 'is lower than']:
                    comp_operator.append(ast.Lt())
                elif self.operator[i] in ['<=', 'is lower or equal to']:
                    comp_operator.append(ast.LtE())
                elif self.operator[i] in ['>', 'is greater than']:
                    comp_operator.append(ast.Gt())
                elif self.operator[i] in ['>=', 'is greater or equal to']:
                    comp_operator.append(ast.GtE())
                elif self.operator[i] in ['==', 'is equal to']:
                    comp_operator.append(ast.Eq())
                elif self.operator[i] in ['!=', 'is different from',
                                          'is not equal to']:
                    comp_operator.append(ast.NotEq())
                elif self.operator[i] == 'in':
                    comp_operator.append(ast.In())
                elif self.operator[i] == 'not in':
                    comp_operator.append(ast.NotIn())
                elif self.operator[i] == 'is':
                    comp_operator.append(ast.Is())
                elif self.operator[i] == 'is not':
                    comp_operator.append(ast.IsNot())
                else:
                    raise Exception("Unrecognized argument in Comparison")
            return ast.Compare(left=self.left.to_node(), ops=comp_operator,
                               comparators=right_nodes)


class Expr(BinaryOp):
    def to_node(self):
        node = self.left.to_node()
        if len(self.operator) == 0:
            return node
        else:
            for right_node in self.right:
                node = ast.BinOp(node, ast.BitOr(), right_node.to_node())
            return node


class XorExpr(BinaryOp):
    def to_node(self):
        node = self.left.to_node()
        if len(self.operator) == 0:
            return node
        else:
            for right_node in self.right:
                node = ast.BinOp(node, ast.BitXor(), right_node.to_node())
            return node


class AndExpr(BinaryOp):
    def to_node(self):
        node = self.left.to_node()
        if len(self.operator) == 0:
            return node
        else:
            for right_node in self.right:
                node = ast.BinOp(node, ast.BitAnd(), right_node.to_node())
            return node


class ShiftExpr(BinaryOp):
    def to_node(self):
        node = self.left.to_node()
        if len(self.operator) == 0:
            return node
        else:
            for right_node in self.right:
                if self.operator[0] == '<<':
                    node = ast.BinOp(node, ast.LShift(), right_node.to_node())
                else:
                    node = ast.BinOp(node, ast.RShift(), right_node.to_node())
            return node


class ArithExpr(BinaryOp):
    def to_node(self):
        """
        Creates a node that represent arithmetic operations (+ and -).
        This node can be inserted into the AST being constructed.
        :return: The node to be inserted into the AST
        """
        node = self.left.to_node()
        if len(self.operator) == 0:
            return node
        else:
            # Because there can be chained operations, the trick is to
            # recursively build the node by creating it as follow:
            # node = ast.BinOp(node, ast.Op(), right) <-- note that node
            # inside ast.BinOp() is reused when creating the final node
            for i in range(len(self.right)):
                if self.operator[i] in ['+', 'plus']:
                    node = ast.BinOp(node, ast.Add(), self.right[i].to_node())
                elif self.operator[i] in ['-', 'minus']:
                    node = ast.BinOp(node, ast.Sub(), self.right[i].to_node())
            return node


class Term(BinaryOp):
    def to_node(self):
        node = self.left.to_node()
        if len(self.operator) == 0:
            return node
        else:
            for i in range(len(self.right)):
                if self.operator[i] in ['*', 'times']:
                    node = ast.BinOp(node, ast.Mult(), self.right[i].to_node())
                elif self.operator[i] in ['/', 'divided by']:
                    node = ast.BinOp(node, ast.Div(), self.right[i].to_node())
                elif self.operator[i] in ['%', 'modulo']:
                    node = ast.BinOp(node, ast.Mod(), self.right[i].to_node())
                elif self.operator[i] == '@':
                    node = ast.BinOp(node, ast.MatMult(),
                                     self.right[i].to_node())
                elif self.operator[i] == '//':
                    node = ast.BinOp(node, ast.FloorDiv(),
                                     self.right[i].to_node())
            return node


class Power(BinaryOp):
    def to_node(self):
        if len(self.operator) == 0:
            return self.left.to_node()
        else:
            return ast.BinOp(self.left.to_node(), ast.Pow(),
                             self.right.to_node())


# -----------------------------------------------------------------------------
# Generators for Unary operators
# -----------------------------------------------------------------------------
class UnaryOp(object):
    def __init__(self, parent, operator, value):
        self.parent = parent
        self.operator = operator
        self.value = value

    def to_node(self):
        pass


class Factor(UnaryOp):
    def to_node(self):
        value = self.value
        operator = self.operator
        if operator in ('+', 'plus'):
            return ast.UnaryOp(ast.UAdd(), value.to_node())
        elif operator in ('-', 'minus'):
            return ast.UnaryOp(ast.USub(), value.to_node())
        elif operator == '~':
            return ast.UnaryOp(ast.Invert(), value.to_node())
        else:
            return value.to_node()


class NotTest(UnaryOp):
    def to_node(self):
        value = self.value
        operator = self.operator
        if operator == 'not':
            return ast.UnaryOp(ast.Not(), value.to_node())
        else:
            return value.to_node()


# -----------------------------------------------------------------------------
# Functions and data structures
# -----------------------------------------------------------------------------
class TestListStarExpr(object):
    def __init__(self, parent, values):
        self.parent = parent
        self.values = values

    def to_node(self):
        """
        This method will be called when both evaluating names or values.
        """
        # If single identifier: list of size 1
        # If multiple identifiers: list of size n
        if len(self.values) > 1:
            elements = []
            for element in self.values:
                elements.append(element.to_node())
            return ast.Tuple(elts=elements, ctx=ast.Load)
        else:
            return self.values[0].to_node()


class TestList(object):
    def __init__(self, parent, range_params, args):
        self.parent = parent
        self.range_params = range_params
        self.args = args

    def to_node(self):
        if self.range_params is not None:
            range_params = []
            for param in self.range_params.args:
                range_params.append(param.to_node())
            node = ast.Call(func=ast.Name(id='range', ctx='Load'),
                            args=range_params, keywords=[])
            return node
        else:
            if len(self.args) == 1:
                return self.args[0].to_node()
            else:
                node = []
                for arg in self.args:
                    node.append(arg.to_node())
                return node


class AtomExpr(object):
    def __init__(self, parent, atom, trailer):
        self.parent = parent
        self.atom = atom
        self.trailer = trailer

    def to_node(self):
        atom = self.atom
        # We are not dealing with Trailers yet, so we comment out the next line
        # trailer = self.trailer
        return atom.to_node()


class Atom(object):
    def __init__(self, parent, is_tuple, is_list, is_dict, values, name,
                 number, string, none, boolean):
        self.parent = parent
        self.is_tuple = is_tuple
        self.is_list = is_list
        self.is_dict = is_dict
        self.values = values
        self.name = name
        self.number = number
        self.string = string
        self.none = none
        self.boolean = boolean

    def to_node(self):
        if self.is_tuple != '':
            elements = self.__create_elements()
            return ast.Tuple(elts=elements, ctx=ast.Load)
        elif self.is_list != '':
            elements = self.__create_elements()
            return ast.List(elts=elements, ctx=ast.Load)
        elif self.is_dict != '':
            if self.values is None:
                return ast.Dict(keys=[], values=[])
            elif self.__is_dict():
                keys, values = self.__create_keys_values()
                return ast.Dict(keys=keys, values=values)
            else:
                elements = self.__create_elements()
                return ast.Set(elts=elements, ctx=ast.Load)
        elif self.name is not None:
            return self.name.to_node()
        elif self.number is not None:
            return self.number.to_node()
        elif self.none is not None:
            return self.none.to_node()
        elif self.boolean is not None:
            return self.boolean.to_node()
        elif self.string is not None:
            return self.string[0].to_node()

    def __create_elements(self):
        elements = []
        if self.values is None:
            return elements
        else:
            values = self.values.values
            for element in values:
                elements.append(element.to_node())
            return elements

    def __is_dict(self):
        return len(self.values.keys) > 0

    def __create_keys_values(self):
        keys = []
        values = []
        for key in self.values.keys:
            keys.append(key.to_node())
        for value in self.values.values:
            values.append(value.to_node())
        return keys, values


class TestListComp(object):
    def __init__(self, parent, values):
        self.parent = parent
        self.values = values

    def to_node(self):
        if len(self.values) > 1:
            elements = []
            for element in self.values:
                elements.append(element.to_node())
            return ast.Tuple(elts=elements, ctx=ast.Load)
        else:
            return self.values.to_node()


class Parameters(object):
    def __init__(self, parent, value):
        self.parent = parent
        self.value = value

    def to_node(self):
        if self.value is not None:
            return self.value.to_node()
        return ast.arguments(args=[], defaults=[], kw_defaults=[], kwarg=None,
                             kwonlyargs=[], vararg=None)


class TypedArgsList(object):
    def __init__(self, parent, type_def, value):
        self.parent = parent
        self.type_def = type_def
        self.value = value

    def to_node(self):
        args = []
        for arg in self.type_def:
            args.append(ast.arg(arg=arg.name.to_node().id, annotation=None))
        defaults = []
        for default in self.value:
            defaults.append(default.to_node())
        return ast.arguments(args=args, defaults=defaults, vararg=None,
                             kwarg=None, kw_default=[], kwonlyargs=[])



import ast

from textx import metamodel_from_file

from pseutopy.generators.expressions import Factor, UnaryOp, NotTest, \
    BinaryOp, OrTest, AndTest, Comparison, TestList, Expr, XorExpr, AndExpr, \
    ShiftExpr, ArithExpr, Term, Power, TestListStarExpr, AtomExpr, Atom, \
    Parameters, TypedArgsList
from pseutopy.generators.statements import Statement, ExprStmt, \
    InputStmt, FuncCallStmt, DeclareStmt, PrintStmt, DelStmt, IfStmt, \
    WhileStmt, ForStmt, FuncDef, ReturnStmt
from pseutopy.generators.values import Number, Name, String, NoneType, \
    Boolean


class PseuToPy(object):
    def __init__(self):
        self.python_ast = ast.Module(body=[])
        self.variables = []
        self.meta_model = metamodel_from_file('./pseutopy/pseudocode.tx',
                                              debug=False,
                                              classes={DeclareStmt, InputStmt,
                                                       PrintStmt, DelStmt,
                                                       FuncCallStmt, IfStmt,
                                                       WhileStmt, ForStmt,
                                                       FuncDef, Parameters,
                                                       TypedArgsList,
                                                       ReturnStmt,
                                                       BinaryOp, OrTest,
                                                       AndTest, Comparison,
                                                       Expr, XorExpr,
                                                       AndExpr, ShiftExpr,
                                                       ArithExpr, Term, Power,
                                                       UnaryOp, Factor,
                                                       NotTest, Statement,
                                                       ExprStmt,
                                                       TestListStarExpr,
                                                       TestList, AtomExpr,
                                                       Atom,
                                                       Number, Name, String,
                                                       NoneType, Boolean})
        self.meta_model.register_obj_processors({
            'Stmt': self.convert,
        })

    def convert_from_file(self, file_name):
        self.reset_ast()
        self.meta_model.model_from_file(file_name)
        return self.python_ast

    def convert_from_string(self, pseudocode_string):
        self.reset_ast()
        self.meta_model.model_from_str(pseudocode_string)
        return self.python_ast

    def reset_ast(self):
        self.python_ast = ast.Module(body=[])
        self.variables = []

    def convert(self, root_statement):
        self.add_to_ast(self.to_node(root_statement))

    def add_to_ast(self, node):
        self.python_ast.body.append(node)
        return node

    def to_node(self, stmt):
        if isinstance(stmt, Statement):
            return stmt.to_node()
        if stmt == 'pass':
            return ast.Pass()
        if stmt == 'break':
            return ast.Break()
        if stmt == 'continue':
            return ast.Continue()



import ast


class Number(object):
    def __init__(self, parent, value):
        self.parent = parent
        self.value = value

    def to_node(self):
        return ast.Num(n=self.value)


class Name(object):
    def __init__(self, parent, id):
        self.parent = parent
        self.id = id

    def to_node(self):
        return ast.Name(id="" + self.id, ctx=ast.Store)


class String(object):
    def __init__(self, parent, value):
        self.parent = parent
        self.value = value

    def to_node(self):
        return ast.Str(s=self.value)


class Boolean(object):
    def __init__(self, parent, value):
        self.parent = parent
        self.value = value

    def to_node(self):
        return ast.NameConstant(value=self.value)


class NoneType(object):
    def __init__(self, parent, value):
        self.parent = parent
        self.value = value

    def to_node(self):
        return ast.NameConstant(value=None)



import ast


class Statement(object):
    def __init__(self, parent):
        self.parent = parent

    def to_node(self):
        pass


# -----------------------------------------------------------------------------
# Generators for simple statements
# -----------------------------------------------------------------------------
class ExprStmt(Statement):
    def __init__(self, parent, name, chained_value, value):
        super().__init__(parent)
        self.name = name
        self.chained_value = chained_value
        self.value = value

    def to_node(self):
        if len(self.chained_value) > 0:
            return self.__create_chained_assignment_node()
        if self.value is not None:
            return self.__create_function_assignment_node()

    def __create_chained_assignment_node(self):
        """
        When creating a chained assignment, it is either:
            - A true chained assignment (e.g., a = b = True)
            - A single assignment (e.g., a = True)
        This method takes care of both these cases
        :return: The node representing a chained assignment
        """
        chained_value = self.chained_value
        value = None
        targets = self.__create_targets()
        for i in range(len(chained_value)):
            if i == len(chained_value) - 1:
                value = chained_value[i].to_node()
            else:
                targets.append(chained_value[i].to_node())
        return ast.Assign(targets=targets, value=value)

    def __create_function_assignment_node(self):
        targets = self.__create_targets()
        return ast.Assign(targets=targets, value=self.value.to_node())

    def __create_targets(self):
        targets = []
        if len(self.name.values) == 1:
            targets.append(ast.Name(id=self.name.to_node(), ctx=ast.Store))
        else:
            elements = []
            for element in self.name.values:
                elements.append(ast.Name(id=element.to_node(), ctx=ast.Store))
            targets.append(ast.Tuple(elts=elements, ctx=ast.Load))
        return targets


class InputStmt(Statement):
    def __init__(self, parent, cast_type, args, values):
        super().__init__(parent)
        self.cast_type = cast_type
        self.args = args
        self.values = values

    def to_node(self):
        args = [self.args.to_node()]
        for value in self.values:
            args.append(value.to_node())
        node = ast.Call(func=ast.Name(id='input', ctx=ast.Load), args=args,
                        keywords=[])
        if self.cast_type == 'number':
            node = ast.Call(func=ast.Name(id='float', ctx=ast.Load),
                            args=[node], keywords=[])
        elif self.cast_type == 'integer':
            node = ast.Call(func=ast.Name(id='int', ctx=ast.Load),
                            args=[node], keywords=[])
        if isinstance(self.parent, ExprStmt):
            return node
        else:
            return ast.Expr(value=node)


class PrintStmt(Statement):
    def __init__(self, parent, args):
        super().__init__(parent)
        self.args = args

    def to_node(self):
        args = []
        for arg in self.args:
            args.append(arg.to_node())
        return ast.Expr(value=ast.Call(func=ast.Name(id='print', ctx=ast.Load),
                                       args=args, keywords=[]))


class FuncCallStmt(Statement):
    def __init__(self, parent, name, args):
        super().__init__(parent)
        self.name = name
        self.args = args

    def to_node(self):
        args = []
        for arg in self.args:
            args.append(arg.to_node())
        if isinstance(self.parent, ExprStmt):
            return ast.Call(
                func=ast.Name(id=self.name.to_node(), ctx=ast.Load),
                args=args, keywords=[])
        else:
            return ast.Expr(value=ast.Call(
                func=ast.Name(id=self.name.to_node(), ctx=ast.Load),
                args=args, keywords=[]))


class DeclareStmt(Statement):
    def __init__(self, parent, name):
        super().__init__(parent)
        self.name = name

    def to_node(self):
        return ast.Assign(targets=[self.name.to_node()],
                          value=ast.NameConstant(value=None))


class DelStmt(Statement):
    def __init__(self, parent, name):
        super().__init__(parent)
        self.name = name

    def to_node(self):
        targets = []
        for name in self.name.value:
            targets.append(name.to_node())
        return ast.Delete(targets=targets)


# -----------------------------------------------------------------------------
# Generators for Compound Statements
# -----------------------------------------------------------------------------
class IfStmt(Statement):
    def __init__(self, parent, condition, body):
        super().__init__(parent)
        self.condition = condition
        self.body = body

    def to_node(self):
        # Because __recursive_orelse function returns a List of size 1,
        # we need to retrieve the first element of the returned result
        return self.__recursive_orelse()[0]

    def __recursive_orelse(self):
        """
        This recursive function creates a AST node for any IfStmt. To make
        the recursion work, we must return the final AST node inside a List.
        This is why we also retrieve the element at index 0 in the to_node()
        function.
        :return: A List of size 1 which contains the whole If statement
        """
        statements = self.body[0].statement
        body = []
        for statement in statements:
            body.append(statement.to_node())
        # This is a case of a single If statement
        if len(self.condition) == 1 and len(self.body) == 1:
            return [ast.If(test=self.condition[0].to_node(), body=body,
                           orelse=[])]
        # This is the case of a 'Else' statement
        elif len(self.condition) == 0 and len(self.body) == 1:
            return body
        # This is a case of a If statement followed by any other statement
        else:
            self.body.pop(0)
            return [ast.If(test=self.condition.pop(0).to_node(), body=body,
                           orelse=self.__recursive_orelse())]


class WhileStmt(Statement):
    def __init__(self, parent, condition, body, else_body):
        super().__init__(parent)
        self.condition = condition
        self.body = body
        self.else_body = else_body

    def to_node(self):
        test = self.condition.to_node()
        body = []
        for statement in self.body.statement:
            body.append(statement.to_node())
        orelse = []
        if self.else_body is not None:
            for statement in self.else_body.statement:
                orelse.append(statement.to_node())
        return ast.While(test=test, body=body, orelse=orelse)


class ForStmt(Statement):
    def __init__(self, parent, target, iterable, body, else_body):
        super().__init__(parent)
        self.target = target
        self.iterable = iterable
        self.body = body
        self.else_body = else_body

    def to_node(self):
        target = self.target.value[0].to_node()
        iterable = self.iterable.to_node()
        body = []
        for statement in self.body.statement:
            body.append(statement.to_node())
        orelse = []
        if self.else_body is not None:
            for statement in self.else_body.statement:
                orelse.append(statement.to_node())
        return ast.For(target=target, iter=iterable, body=body, orelse=orelse)


class FuncDef(Statement):
    def __init__(self, parent, name, params, test, body):
        super().__init__(parent)
        self.name = name
        self.params = params
        self.test = test
        self.body = body

    def to_node(self):
        name = self.name.id
        if self.params is not None:
            args = self.params.to_node()
        else:
            args = ast.arguments(args=[], defaults=[], kw_defaults=[],
                                 kwarg=None, kwonlyargs=[], vararg=None)
        body = []
        for statement in self.body.stmt:
            body.append(statement.to_node())
        return ast.FunctionDef(name=name, args=args, body=body,
                               decorator_list=[])


class ReturnStmt(Statement):
    def __init__(self, parent, value):
        super().__init__(parent)
        self.value = value

    def to_node(self):
        return ast.Return(value=self.value.to_node())
